@@ -1,4 +1,4 @@
-# Copyright 2013-2015 DataStax, Inc.
+# Copyright 2013-2017 DataStax, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -43,16 +43,21 @@ import sys
 from uuid import UUID
 import warnings
 
+if six.PY3:
+    import ipaddress
 
 from cassandra.marshal import (int8_pack, int8_unpack, int16_pack, int16_unpack,
                                uint16_pack, uint16_unpack, uint32_pack, uint32_unpack,
                                int32_pack, int32_unpack, int64_pack, int64_unpack,
                                float_pack, float_unpack, double_pack, double_unpack,
-                               varint_pack, varint_unpack)
+                               varint_pack, varint_unpack, vints_pack, vints_unpack)
 from cassandra import util
 
 apache_cassandra_type_prefix = 'org.apache.cassandra.db.marshal.'
 
+cassandra_empty_type = 'org.apache.cassandra.db.marshal.EmptyType'
+cql_empty_type = 'empty'
+
 log = logging.getLogger(__name__)
 
 if six.PY3:
@@ -73,21 +78,19 @@ def trim_if_startswith(s, prefix):
     return s
 
 
-def unix_time_from_uuid1(u):
-    msg = "'cassandra.cqltypes.unix_time_from_uuid1' has moved to 'cassandra.util'. This entry point will be removed in the next major version."
-    warnings.warn(msg, DeprecationWarning)
-    log.warning(msg)
-    return util.unix_time_from_uuid1(u)
-
-
-def datetime_from_timestamp(timestamp):
-    msg = "'cassandra.cqltypes.datetime_from_timestamp' has moved to 'cassandra.util'. This entry point will be removed in the next major version."
-    warnings.warn(msg, DeprecationWarning)
-    log.warning(msg)
-    return util.datetime_from_timestamp(timestamp)
-
-
 _casstypes = {}
+_cqltypes = {}
+
+
+cql_type_scanner = re.Scanner((
+    ('frozen', None),
+    (r'[a-zA-Z0-9_]+', lambda s, t: t),
+    (r'[\s,<>]', None),
+))
+
+
+def cql_types_from_string(cql_type):
+    return cql_type_scanner.scan(cql_type)[0]
 
 
 class CassandraTypeType(type):
@@ -106,6 +109,8 @@ class CassandraTypeType(type):
         cls = type.__new__(metacls, name, bases, dct)
         if not name.startswith('_'):
             _casstypes[name] = cls
+            if not cls.typename.startswith(apache_cassandra_type_prefix):
+                _cqltypes[cls.typename] = cls
         return cls
 
 
@@ -173,7 +178,7 @@ def lookup_casstype(casstype):
     Example:
 
         >>> lookup_casstype('org.apache.cassandra.db.marshal.MapType(org.apache.cassandra.db.marshal.UTF8Type,org.apache.cassandra.db.marshal.Int32Type)')
-        <class 'cassandra.types.MapType(UTF8Type, Int32Type)'>
+        <class 'cassandra.cqltypes.MapType(UTF8Type, Int32Type)'>
 
     """
     if isinstance(casstype, (CassandraType, CassandraTypeType)):
@@ -184,6 +189,10 @@ def lookup_casstype(casstype):
         raise ValueError("Don't know how to parse type string %r: %s" % (casstype, e))
 
 
+def is_reversed_casstype(data_type):
+    return issubclass(data_type, ReversedType)
+
+
 class EmptyValue(object):
     """ See _CassandraType.support_empty_values """
 
@@ -289,7 +298,7 @@ class _CassandraType(object):
         using them as parameters. This is how composite types are constructed.
 
             >>> MapType.apply_parameters([DateType, BooleanType])
-            <class 'cassandra.types.MapType(DateType, BooleanType)'>
+            <class 'cassandra.cqltypes.MapType(DateType, BooleanType)'>
 
         `subtypes` will be a sequence of CassandraTypes.  If provided, `names`
         will be an equally long sequence of column names or Nones.
@@ -365,7 +374,10 @@ class DecimalType(_CassandraType):
         try:
             sign, digits, exponent = dec.as_tuple()
         except AttributeError:
-            raise TypeError("Non-Decimal type received for Decimal value")
+            try:
+                sign, digits, exponent = Decimal(dec).as_tuple()
+            except Exception:
+                raise TypeError("Invalid type for Decimal value: %r", dec)
         unscaled = int(''.join([str(digit) for digit in digits]))
         if sign:
             unscaled *= -1
@@ -507,12 +519,17 @@ class InetAddressType(_CassandraType):
 
     @staticmethod
     def serialize(addr, protocol_version):
-        if ':' in addr:
-            return util.inet_pton(socket.AF_INET6, addr)
-        else:
-            # util.inet_pton could also handle, but this is faster
-            # since we've already determined the AF
-            return socket.inet_aton(addr)
+        try:
+            if ':' in addr:
+                return util.inet_pton(socket.AF_INET6, addr)
+            else:
+                # util.inet_pton could also handle, but this is faster
+                # since we've already determined the AF
+                return socket.inet_aton(addr)
+        except:
+            if six.PY3 and isinstance(addr, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
+                return addr.packed
+            raise ValueError("can't interpret %r as an inet address" % (addr,))
 
 
 class CounterColumnType(LongType):
@@ -534,7 +551,6 @@ class DateType(_CassandraType):
 
     @staticmethod
     def interpret_datestring(val):
-        # not used internally. deprecate?
         if val[-5] in ('+', '-'):
             offset = (int(val[-4:-2]) * 3600 + int(val[-2:]) * 60) * int(val[-5] + '1')
             val = val[:-5]
@@ -562,10 +578,13 @@ class DateType(_CassandraType):
             timestamp_seconds = calendar.timegm(v.utctimetuple())
             timestamp = timestamp_seconds * 1e3 + getattr(v, 'microsecond', 0) / 1e3
         except AttributeError:
-            # Ints and floats are valid timestamps too
-            if type(v) not in _number_types:
-                raise TypeError('DateType arguments must be a datetime or timestamp')
-            timestamp = v
+            try:
+                timestamp = calendar.timegm(v.timetuple()) * 1e3
+            except AttributeError:
+                # Ints and floats are valid timestamps too
+                if type(v) not in _number_types:
+                    raise TypeError('DateType arguments must be a datetime, date, or timestamp')
+                timestamp = v
 
         return int64_pack(long(timestamp))
 
@@ -578,7 +597,7 @@ class TimeUUIDType(DateType):
     typename = 'timeuuid'
 
     def my_timestamp(self):
-        return unix_time_from_uuid1(self.val)
+        return util.unix_time_from_uuid1(self.val)
 
     @staticmethod
     def deserialize(byts, protocol_version):
@@ -611,6 +630,11 @@ class SimpleDateType(_CassandraType):
         try:
             days = val.days_from_epoch
         except AttributeError:
+            if isinstance(val, six.integer_types):
+                # the DB wants offset int values, but util.Date init takes days from epoch
+                # here we assume int values are offset, as they would appear in CQL
+                # short circuit to avoid subtracting just to add offset
+                return uint32_pack(val)
             days = util.Date(val).days_from_epoch
         return uint32_pack(days + SimpleDateType.EPOCH_OFFSET_DAYS)
 
@@ -643,6 +667,23 @@ class TimeType(_CassandraType):
         return int64_pack(nano)
 
 
+class DurationType(_CassandraType):
+    typename = 'duration'
+
+    @staticmethod
+    def deserialize(byts, protocol_version):
+        months, days, nanoseconds = vints_unpack(byts)
+        return util.Duration(months, days, nanoseconds)
+
+    @staticmethod
+    def serialize(duration, protocol_version):
+        try:
+            m, d, n = duration.months, duration.days, duration.nanoseconds
+        except AttributeError:
+            raise TypeError('DurationType arguments must be a Duration.')
+        return vints_pack([m, d, n])
+
+
 class UTF8Type(_CassandraType):
     typename = 'text'
     empty_binary_ok = True
@@ -665,6 +706,8 @@ class VarcharType(UTF8Type):
 
 
 class _ParameterizedType(_CassandraType):
+    num_subtypes = 'UNKNOWN'
+
     @classmethod
     def deserialize(cls, byts, protocol_version):
         if not cls.subtypes:
@@ -693,12 +736,13 @@ class _SimpleParameterizedType(_ParameterizedType):
         numelements = unpack(byts[:length])
         p = length
         result = []
+        inner_proto = max(3, protocol_version)
         for _ in range(numelements):
             itemlen = unpack(byts[p:p + length])
             p += length
             item = byts[p:p + itemlen]
             p += itemlen
-            result.append(subtype.from_binary(item, protocol_version))
+            result.append(subtype.from_binary(item, inner_proto))
         return cls.adapter(result)
 
     @classmethod
@@ -710,8 +754,9 @@ class _SimpleParameterizedType(_ParameterizedType):
         pack = int32_pack if protocol_version >= 3 else uint16_pack
         buf = io.BytesIO()
         buf.write(pack(len(items)))
+        inner_proto = max(3, protocol_version)
         for item in items:
-            itembytes = subtype.to_binary(item, protocol_version)
+            itembytes = subtype.to_binary(item, inner_proto)
             buf.write(pack(len(itembytes)))
             buf.write(itembytes)
         return buf.getvalue()
@@ -745,6 +790,7 @@ class MapType(_ParameterizedType):
         numelements = unpack(byts[:length])
         p = length
         themap = util.OrderedMapSerializedKey(key_type, protocol_version)
+        inner_proto = max(3, protocol_version)
         for _ in range(numelements):
             key_len = unpack(byts[p:p + length])
             p += length
@@ -754,8 +800,8 @@ class MapType(_ParameterizedType):
             p += length
             valbytes = byts[p:p + val_len]
             p += val_len
-            key = key_type.from_binary(keybytes, protocol_version)
-            val = value_type.from_binary(valbytes, protocol_version)
+            key = key_type.from_binary(keybytes, inner_proto)
+            val = value_type.from_binary(valbytes, inner_proto)
             themap._insert_unchecked(key, keybytes, val)
         return themap
 
@@ -769,9 +815,10 @@ class MapType(_ParameterizedType):
             items = six.iteritems(themap)
         except AttributeError:
             raise TypeError("Got a non-map object for a map value")
+        inner_proto = max(3, protocol_version)
         for key, val in items:
-            keybytes = key_type.to_binary(key, protocol_version)
-            valbytes = value_type.to_binary(val, protocol_version)
+            keybytes = key_type.to_binary(key, inner_proto)
+            valbytes = value_type.to_binary(val, inner_proto)
             buf.write(pack(len(keybytes)))
             buf.write(keybytes)
             buf.write(pack(len(valbytes)))
@@ -781,7 +828,6 @@ class MapType(_ParameterizedType):
 
 class TupleType(_ParameterizedType):
     typename = 'tuple'
-    num_subtypes = 'UNKNOWN'
 
     @classmethod
     def deserialize_safe(cls, byts, protocol_version):
@@ -832,28 +878,29 @@ class TupleType(_ParameterizedType):
 
 
 class UserType(TupleType):
-    typename = "'org.apache.cassandra.db.marshal.UserType'"
+    typename = "org.apache.cassandra.db.marshal.UserType"
 
     _cache = {}
     _module = sys.modules[__name__]
 
     @classmethod
-    def make_udt_class(cls, keyspace, udt_name, names_and_types, mapped_class):
+    def make_udt_class(cls, keyspace, udt_name, field_names, field_types):
+        assert len(field_names) == len(field_types)
+
         if six.PY2 and isinstance(udt_name, unicode):
             udt_name = udt_name.encode('utf-8')
-        try:
-            return cls._cache[(keyspace, udt_name)]
-        except KeyError:
-            field_names, types = zip(*names_and_types)
-            instance = type(udt_name, (cls,), {'subtypes': types,
+
+        instance = cls._cache.get((keyspace, udt_name))
+        if not instance or instance.fieldnames != field_names or instance.subtypes != field_types:
+            instance = type(udt_name, (cls,), {'subtypes': field_types,
                                                'cassname': cls.cassname,
                                                'typename': udt_name,
                                                'fieldnames': field_names,
                                                'keyspace': keyspace,
-                                               'mapped_class': mapped_class,
+                                               'mapped_class': None,
                                                'tuple_type': cls._make_registered_udt_namedtuple(keyspace, udt_name, field_names)})
             cls._cache[(keyspace, udt_name)] = instance
-            return instance
+        return instance
 
     @classmethod
     def evict_udt_class(cls, keyspace, udt_name):
@@ -866,17 +913,10 @@ class UserType(TupleType):
 
     @classmethod
     def apply_parameters(cls, subtypes, names):
-        keyspace = subtypes[0]
+        keyspace = subtypes[0].cass_parameterized_type()  # when parsed from cassandra type, the keyspace is created as an unrecognized cass type; This gets the name back
         udt_name = _name_from_hex_string(subtypes[1].cassname)
-        field_names = [_name_from_hex_string(encoded_name) for encoded_name in names[2:]]
-        assert len(field_names) == len(subtypes[2:])
-        return type(udt_name, (cls,), {'subtypes': subtypes[2:],
-                                       'cassname': cls.cassname,
-                                       'typename': udt_name,
-                                       'fieldnames': field_names,
-                                       'keyspace': keyspace,
-                                       'mapped_class': None,
-                                       'tuple_type': namedtuple(udt_name, field_names)})
+        field_names = tuple(_name_from_hex_string(encoded_name) for encoded_name in names[2:])  # using tuple here to match what comes into make_udt_class from other sources (for caching equality test)
+        return cls.make_udt_class(keyspace, udt_name, field_names, tuple(subtypes[2:]))
 
     @classmethod
     def cql_parameterized_type(cls):
@@ -887,17 +927,24 @@ class UserType(TupleType):
         values = super(UserType, cls).deserialize_safe(byts, protocol_version)
         if cls.mapped_class:
             return cls.mapped_class(**dict(zip(cls.fieldnames, values)))
-        else:
+        elif cls.tuple_type:
             return cls.tuple_type(*values)
+        else:
+            return tuple(values)
 
     @classmethod
     def serialize_safe(cls, val, protocol_version):
         proto_version = max(3, protocol_version)
         buf = io.BytesIO()
-        for fieldname, subtype in zip(cls.fieldnames, cls.subtypes):
-            item = getattr(val, fieldname)
+        for i, (fieldname, subtype) in enumerate(zip(cls.fieldnames, cls.subtypes)):
+            # first treat as a tuple, else by custom type
+            try:
+                item = val[i]
+            except TypeError:
+                item = getattr(val, fieldname)
+
             if item is not None:
-                packed_item = subtype.to_binary(getattr(val, fieldname), proto_version)
+                packed_item = subtype.to_binary(item, proto_version)
                 buf.write(int32_pack(len(packed_item)))
                 buf.write(packed_item)
             else:
@@ -909,17 +956,32 @@ class UserType(TupleType):
         # this is required to make the type resolvable via this module...
         # required when unregistered udts are pickled for use as keys in
         # util.OrderedMap
-        qualified_name = "%s_%s" % (keyspace, name)
-        nt = getattr(cls._module, qualified_name, None)
-        if not nt:
-            nt = namedtuple(qualified_name, field_names)
-            setattr(cls._module, qualified_name, nt)
-        return nt
+        t = cls._make_udt_tuple_type(name, field_names)
+        if t:
+            qualified_name = "%s_%s" % (keyspace, name)
+            setattr(cls._module, qualified_name, t)
+        return t
+
+    @classmethod
+    def _make_udt_tuple_type(cls, name, field_names):
+        # fallback to positional named, then unnamed tuples
+        # for CQL identifiers that aren't valid in Python,
+        try:
+            t = namedtuple(name, field_names)
+        except ValueError:
+            try:
+                t = namedtuple(name, util._positional_rename_invalid_identifiers(field_names))
+                log.warn("could not create a namedtuple for '%s' because one or more field names are not valid Python identifiers (%s); " \
+                         "returning positionally-named fields" % (name, field_names))
+            except ValueError:
+                t = None
+                log.warn("could not create a namedtuple for '%s' because the name is not a valid Python identifier; " \
+                         "will return tuples in its place" % (name,))
+        return t
 
 
 class CompositeType(_ParameterizedType):
-    typename = "'org.apache.cassandra.db.marshal.CompositeType'"
-    num_subtypes = 'UNKNOWN'
+    typename = "org.apache.cassandra.db.marshal.CompositeType"
 
     @classmethod
     def cql_parameterized_type(cls):
@@ -947,8 +1009,13 @@ class CompositeType(_ParameterizedType):
         return tuple(result)
 
 
-class DynamicCompositeType(CompositeType):
-    typename = "'org.apache.cassandra.db.marshal.DynamicCompositeType'"
+class DynamicCompositeType(_ParameterizedType):
+    typename = "org.apache.cassandra.db.marshal.DynamicCompositeType"
+
+    @classmethod
+    def cql_parameterized_type(cls):
+        sublist = ', '.join('%s=>%s' % (alias, typ.cass_parameterized_type(full=True)) for alias, typ in zip(cls.fieldnames, cls.subtypes))
+        return "'%s(%s)'" % (cls.typename, sublist)
 
 
 class ColumnToCollectionType(_ParameterizedType):
@@ -957,18 +1024,17 @@ class ColumnToCollectionType(_ParameterizedType):
     Cassandra includes this. We don't actually need or want the extra
     information.
     """
-    typename = "'org.apache.cassandra.db.marshal.ColumnToCollectionType'"
-    num_subtypes = 'UNKNOWN'
+    typename = "org.apache.cassandra.db.marshal.ColumnToCollectionType"
 
 
 class ReversedType(_ParameterizedType):
-    typename = "'org.apache.cassandra.db.marshal.ReversedType'"
+    typename = "org.apache.cassandra.db.marshal.ReversedType"
     num_subtypes = 1
 
     @classmethod
     def deserialize_safe(cls, byts, protocol_version):
         subtype, = cls.subtypes
-        return subtype.from_binary(byts)
+        return subtype.from_binary(byts, protocol_version)
 
     @classmethod
     def serialize_safe(cls, val, protocol_version):
@@ -983,7 +1049,7 @@ class FrozenType(_ParameterizedType):
     @classmethod
     def deserialize_safe(cls, byts, protocol_version):
         subtype, = cls.subtypes
-        return subtype.from_binary(byts)
+        return subtype.from_binary(byts, protocol_version)
 
     @classmethod
     def serialize_safe(cls, val, protocol_version):
