@@ -56,14 +56,14 @@ fi
 
 # Set up venv with dtest dependencies
 set -e # enable immediate exit if venv setup fails
-virtualenv --python=python2 --no-site-packages venv
+virtualenv --python=python2 venv
 source venv/bin/activate
 pip install -r ${CASSANDRA_DIR}/pylib/requirements.txt
 pip freeze
 
 if [ "$cython" = "yes" ]; then
     TESTSUITE_NAME="${TESTSUITE_NAME}.cython"
-    pip install "Cython>=0.20,<0.25"
+    pip install "Cython>=0.27.2,<0.28"
     cd pylib/; python setup.py build_ext --inplace
     cd ${WORKSPACE}
 else
